# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""\
Fitting transformations --- :mod:`MDAnalysis.transformations.fit`
=================================================================

Translate and/or rotates the coordinates of a given trajectory to align
a given AtomGroup to a reference structure.

.. autoclass:: fit_translation

.. autoclass:: fit_rot_trans

"""
import numpy as np

from ..analysis import align
from ..lib.transformations import euler_from_matrix, euler_matrix

from .base import TransformationBase


class fit_translation(TransformationBase):
    """Translates a given AtomGroup so that its center of geometry/mass matches
    the respective center of the given reference. A plane can be given by the
    user using the option `plane`, and will result in the removal of
    the translation motions of the AtomGroup over that particular plane.

    Example
    -------
    Removing the translations of a given AtomGroup `ag` on the XY plane by
    fitting its center of mass to the center of mass of a reference `ref`:

    .. code-block:: python

        ag = u.select_atoms("protein")
        ref = mda.Universe("reference.pdb")
        transform = mda.transformations.fit_translation(ag, ref, plane="xy",
                                                        weights="mass")
        u.trajectory.add_transformations(transform)

    Parameters
    ----------
    ag : Universe or AtomGroup
       structure to translate, a
       :class:`~MDAnalysis.core.groups.AtomGroup` or a whole
       :class:`~MDAnalysis.core.universe.Universe`
    reference : Universe or AtomGroup
       reference structure, a :class:`~MDAnalysis.core.groups.AtomGroup` or a
       whole :class:`~MDAnalysis.core.universe.Universe`
    plane: str, optional
        used to define the plane on which the translations will be removed.
        Defined as a string of the plane.
        Suported planes are yz, xz and xy planes.
    weights : {"mass", ``None``} or array_like, optional
       choose weights. With ``"mass"`` uses masses as weights; with ``None``
       weigh each atom equally. If a float array of the same length as
       `ag` is provided, use each element of the `array_like` as a
       weight for the corresponding atom in `ag`.

    Returns
    -------
    MDAnalysis.coordinates.timestep.Timestep


    .. versionchanged:: 2.0.0
       The transformation was changed from a function/closure to a class
       with ``__call__``.
    .. versionchanged:: 2.0.0
       The transformation was changed to inherit from the base class for
       limiting threads and checking if it can be used in parallel analysis.
    """

    def __init__(
        self,
        ag,
        reference,
        plane=None,
        weights=None,
        max_threads=None,
        parallelizable=True,
    ):
        super().__init__(
            max_threads=max_threads, parallelizable=parallelizable
        )

        self.ag = ag
        self.reference = reference
        self.plane = plane
        self.weights = weights

        if self.plane is not None:
            axes = {"yz": 0, "xz": 1, "xy": 2}
            try:
                self.plane = axes[self.plane]
            except (TypeError, KeyError):
                raise ValueError(
                    f"{self.plane} is not a valid plane"
                ) from None
        try:
            if self.ag.atoms.n_residues != self.reference.atoms.n_residues:
                errmsg = (
                    f"{self.ag} and {self.reference} have mismatched"
                    f"number of residues"
                )

                raise ValueError(errmsg)
        except AttributeError:
            errmsg = (
                f"{self.ag} or {self.reference} is not valid"
                f"Universe/AtomGroup"
            )
            raise AttributeError(errmsg) from None
        self.ref, self.mobile = align.get_matching_atoms(
            self.reference.atoms, self.ag.atoms
        )
        self.weights = align.get_weights(self.ref.atoms, weights=self.weights)
        self.ref_com = self.ref.center(self.weights)

    def _transform(self, ts):
        mobile_com = np.asarray(
            self.mobile.atoms.center(self.weights), np.float32
        )
        vector = self.ref_com - mobile_com
        if self.plane is not None:
            vector[self.plane] = 0
        ts.positions += vector

        return ts


class fit_rot_trans(TransformationBase):
    """Perform a spatial superposition by minimizing the RMSD.

    Spatially align the group of atoms `ag` to `reference` by doing a RMSD
    fit.

    This fit works as a way to remove translations and rotations of a given
    AtomGroup in a trajectory. A plane can be given using the flag `plane`
    so that only translations and rotations in that particular plane are
    removed. This is useful for protein-membrane systems to where the membrane
    must remain in the same orientation.

    Note
    ----
    ``max_threads`` is set to 1 for this transformation
    with which it performs better.

    Example
    -------
    Removing the translations and rotations of a given AtomGroup `ag` on the XY plane
    by fitting it to a reference `ref`, using the masses as weights for the RMSD fit:

    .. code-block:: python

        ag = u.select_atoms("protein")
        ref = mda.Universe("reference.pdb")
        transform = mda.transformations.fit_rot_trans(ag, ref, plane="xy",
                                                      weights="mass")
        u.trajectory.add_transformations(transform)

    Parameters
    ----------
    ag : Universe or AtomGroup
       structure to translate and rotate, a
       :class:`~MDAnalysis.core.groups.AtomGroup` or a whole
       :class:`~MDAnalysis.core.universe.Universe`
    reference : Universe or AtomGroup
       reference structure, a :class:`~MDAnalysis.core.groups.AtomGroup` or a whole
       :class:`~MDAnalysis.core.universe.Universe`
    plane: str, optional
        used to define the plane on which the rotations and translations will be removed.
        Defined as a string of the plane. Supported planes are "yz", "xz" and "xy" planes.
    weights : {"mass", ``None``} or array_like, optional
       choose weights. With ``"mass"`` uses masses as weights; with ``None``
       weigh each atom equally. If a float array of the same length as
       `ag` is provided, use each element of the `array_like` as a
       weight for the corresponding atom in `ag`.

    Returns
    -------
    MDAnalysis.coordinates.timestep.Timestep


    .. versionchanged:: 2.0.0
       The transformation was changed from a function/closure to a class
       with ``__call__``.
    .. versionchanged:: 2.0.0
       The transformation was changed to inherit from the base class for
       limiting threads and checking if it can be used in parallel analysis.
    """

    def __init__(
        self,
        ag,
        reference,
        plane=None,
        weights=None,
        max_threads=1,
        parallelizable=True,
    ):
        super().__init__(
            max_threads=max_threads, parallelizable=parallelizable
        )

        self.ag = ag
        self.reference = reference
        self.plane = plane
        self.weights = weights

        if self.plane is not None:
            axes = {"yz": 0, "xz": 1, "xy": 2}
            try:
                self.plane = axes[self.plane]
            except (TypeError, KeyError):
                raise ValueError(
                    f"{self.plane} is not a valid plane"
                ) from None
        try:
            if self.ag.atoms.n_residues != self.reference.atoms.n_residues:
                errmsg = (
                    f"{self.ag} and {self.reference} have mismatched "
                    f"number of residues"
                )
                raise ValueError(errmsg)
        except AttributeError:
            errmsg = (
                f"{self.ag} or {self.reference} is not valid "
                f"Universe/AtomGroup"
            )
            raise AttributeError(errmsg) from None
        self.ref, self.mobile = align.get_matching_atoms(
            self.reference.atoms, self.ag.atoms
        )
        self.weights = align.get_weights(self.ref.atoms, weights=self.weights)
        self.ref_com = self.ref.center(self.weights)
        self.ref_coordinates = self.ref.atoms.positions - self.ref_com

    def _transform(self, ts):
        mobile_com = self.mobile.atoms.center(self.weights)
        mobile_coordinates = self.mobile.atoms.positions - mobile_com
        rotation, dump = align.rotation_matrix(
            mobile_coordinates, self.ref_coordinates, weights=self.weights
        )
        vector = self.ref_com
        if self.plane is not None:
            matrix = np.r_[rotation, np.zeros(3).reshape(1, 3)]
            matrix = np.c_[matrix, np.zeros(4)]
            euler_angs = np.asarray(
                euler_from_matrix(matrix, axes="sxyz"), np.float32
            )
            for i in range(0, euler_angs.size):
                euler_angs[i] = (
                    euler_angs[self.plane] if i == self.plane else 0
                )
            rotation = euler_matrix(
                euler_angs[0], euler_angs[1], euler_angs[2], axes="sxyz"
            )[:3, :3]
            vector[self.plane] = mobile_com[self.plane]
        ts.positions = ts.positions - mobile_com
        ts.positions = np.dot(ts.positions, rotation.T)
        ts.positions = ts.positions + vector
        return ts



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
#

# cython: embedsignature=True


"""
Timestep Class --- :mod:`MDAnalysis.coordinates.timestep`
============================================================

Derive other Timestep, classes from the classes in this module.
The derived classes must follow the :ref:`Trajectory API`.

Timestep
--------

A :class:`Timestep` holds information for the current time frame in
the trajectory. It is one of the central data structures in
MDAnalysis.

.. class:: Timestep

   .. automethod:: __init__
   .. automethod:: from_coordinates
   .. automethod:: from_timestep
   .. autoattribute:: n_atoms
   .. attribute:: frame

      frame number (0-based)

      .. versionchanged:: 0.11.0
         Frames now 0-based; was 1-based


   .. autoattribute:: time
   .. autoattribute:: dt
   .. autoattribute:: positions
   .. autoattribute:: velocities
   .. autoattribute:: forces
   .. autoattribute:: has_positions
   .. autoattribute:: has_velocities
   .. autoattribute:: has_forces
   .. attribute:: _pos

      :class:`numpy.ndarray` of dtype :class:`~numpy.float32` of shape
      (*n_atoms*, 3) and internal C order, holding the raw
      cartesian coordinates (in MDAnalysis units, i.e. Å).

      .. Note::

         Normally one does not directly access :attr:`_pos` but uses
         the :attr:`~MDAnalysis.core.groups.AtomGroup.positions`
         attribute of a :class:`~MDAnalysis.core.groups.AtomGroup` but
         sometimes it can be faster to directly use the raw
         coordinates. Any changes to this array are immediately
         reflected in the atom positions. If the frame is written to a new
         trajectory then the coordinates are changed. If a new
         trajectory frame is loaded, then *all* contents of
         :attr:`_pos` are overwritten.

   .. attribute:: _velocities

      :class:`numpy.ndarray` of dtype :class:`~numpy.float32`. of shape
      (*n_atoms*, 3), holding the raw velocities (in MDAnalysis
      units, i.e., Å/ps).

      .. Note::

         Normally velocities are accessed through the
         :attr:`Timestep.velocities` attribute or the
         :attr:`~MDAnalysis.core.groups.AtomGroup.velocities`
         attribute of an :class:`~MDAnalysis.core.groups.AtomGroup`

         :attr:`~Timestep._velocities` only exists if the :attr:`has_velocities`
         flag is ``True``.

      .. versionadded:: 0.7.5

   .. attribute:: _forces

      :class:`numpy.ndarray` of dtype :class:`~numpy.float32`. of shape
      (*n_atoms*, 3), holding the forces

      :attr:`~Timestep._forces` only exists if :attr:`has_forces`
      is ``True``.

      .. versionadded:: 0.11.0
         Added as optional to :class:`Timestep`

   .. autoattribute:: dtype
   .. autoattribute:: dimensions
   .. autoattribute:: triclinic_dimensions
   .. autoattribute:: volume
   .. attribute:: data

      :class:`dict` that holds arbitrary per Timestep data

      .. versionadded:: 0.11.0

   .. automethod:: __getitem__
   .. automethod:: __eq__
   .. automethod:: __iter__
   .. automethod:: copy
   .. automethod:: copy_slice

"""


from ..lib.util import Namespace
from ..exceptions import NoDataError
from . import core
from libc.stdint cimport uint64_t
import weakref
import warnings
import copy
import numbers

import numpy as np
cimport numpy as cnp
cnp.import_array()


cdef class Timestep:
    """Timestep data for one frame

    :Methods:

      ``ts = Timestep(n_atoms)``

         create a timestep object with space for n_atoms


    .. versionchanged:: 0.11.0
       Added :meth:`from_timestep` and :meth:`from_coordinates` constructor
       methods.
       :class:`Timestep` init now only accepts integer creation.
       :attr:`n_atoms` now a read only property.
       :attr:`frame` now 0-based instead of 1-based.
       Attributes `status` and `step` removed.
    .. versionchanged:: 2.0.0
       Timestep now can be (un)pickled. Weakref for Reader
       will be dropped.
       Timestep now stores in to numpy array memory in 'C' order rather than
       'F' (Fortran).
    .. versionchanged:: 2.3.0
        Timestep is now a Cython extension type.
        All arrays are now forced to be C contiguous using the NumPy C API.
    """

    order = 'C'

    def __cinit__(self, uint64_t n_atoms, **kwargs):
        """Initialise C++ level parameters of a Timestep

        Parameters
        ----------
        n_atoms : uint64
          The total number of atoms this Timestep describes


        .. versionadded:: 2.3.0
           Initialise C++ level parameters
        """
        # c++ level objects
        self._n_atoms = n_atoms
        self.frame = -1

        # NOTE  This is currently hardcoded to match MDA always casting to F32
        # meaning the DTYPE set in the args is not respected.
        # to fix remove hardcode with introspection of a dtype kwarg following
        # discussion of appropriate casting rules
        self._typenum = cnp.NPY_FLOAT32

        self._has_positions = False
        self._has_velocities = False
        self._has_forces = False

        # track whether array has been allocated correct size
        self._positions_alloc = False
        self._velocities_alloc = False
        self._forces_alloc = False

        # use this to create numpy zeros and empties of the right shape using
        # NumPy C API
        self._particle_dependent_dim[0] = self._n_atoms
        self._particle_dependent_dim[1] = 3

        # unitcell uses a temp only as its size can vary
        cdef cnp.npy_intp unitcell_dim_tmp[1]
        unitcell_dim_tmp[0] = 6

        # use temps so that we don't have to allocate a bunch of empty
        # arrays of large size, eg for vel and frc
        cdef cnp.npy_intp particle_dependent_dim_tmp[2]
        particle_dependent_dim_tmp[0] = 0
        particle_dependent_dim_tmp[1] = 0

        # these must be initialised, we initialise size 0
        self._unitcell = cnp.PyArray_ZEROS(
            1, unitcell_dim_tmp, self._typenum, 0)
        self._pos = cnp.PyArray_EMPTY(
            2, particle_dependent_dim_tmp, self._typenum, 0)
        self._velocities = cnp.PyArray_EMPTY(
            2, particle_dependent_dim_tmp, self._typenum, 0)
        self._forces = cnp.PyArray_EMPTY(
            2, particle_dependent_dim_tmp, self._typenum, 0)

    def __init__(self, uint64_t n_atoms, **kwargs):
        """Create a Timestep, representing a frame of a trajectory

        Parameters
        ----------
        n_atoms : uint64
          The total number of atoms this Timestep describes
        positions : bool, optional
          Whether this Timestep has position information [``True``]
        velocities : bool (optional)
          Whether this Timestep has velocity information [``False``]
        forces : bool (optional)
          Whether this Timestep has force information [``False``]
        reader : Reader (optional)
          A weak reference to the owning Reader.  Used for
          when attributes require trajectory manipulation (e.g. dt)
        dt : float (optional)
          The time difference between frames (ps).  If :attr:`time`
          is set, then `dt` will be ignored.
        time_offset : float (optional)
          The starting time from which to calculate time (in ps)


        .. versionchanged:: 0.11.0
           Added keywords for `positions`, `velocities` and `forces`.
           Can add and remove position/velocity/force information by using
           the ``has_*`` attribute.
        .. versionchanged:: 2.3.0
           Added the `dtype` attribute hardcoded to :class:`~numpy.float32`.
        """
        # hardcoded
        self._dtype = np.float32

        self.data = {}

        for att in ('dt', 'time_offset'):
            try:
                self.data[att] = kwargs[att]
            except KeyError:
                pass
        try:
            # do I have a hook back to the Reader?
            # can possibly __cinit__ this with PyWeakRefNew
            self._reader = weakref.ref(kwargs['reader'])
        except KeyError:
            pass

        self.has_positions = kwargs.get('positions', True)
        self.has_velocities = kwargs.get('velocities', False)
        self.has_forces = kwargs.get('forces', False)

        # set up aux namespace for adding auxiliary data
        self.aux = Namespace()

    def __dealloc__(self):
        pass

    @property
    def n_atoms(self):
        """A read only view of the number of atoms this Timestep has

        .. versionchanged:: 0.11.0
           Changed to read only property
        """
        # In future could do some magic here to make setting n_atoms
        # resize the coordinate arrays, but
        # - not sure if that is ever useful
        # - not sure how to manage existing data upon extension
        return self._n_atoms

    @property
    def dtype(self):
        """The NumPy dtype of the timestep, all arrays in the timestep will
            have this dtype. Currently hardcoded to :class:`~numpy.float32`.

        .. versionadded:: 2.3.0
           Added dtype
        """
        return self._dtype

    @property
    def has_positions(self):
        """A boolean of whether this Timestep has position data
        This can be changed to ``True`` or ``False`` to allocate space for
        or remove the data.

        .. versionadded:: 0.11.0
        """
        return self._has_positions

    @has_positions.setter
    def has_positions(self, val):
        if val and not self._has_positions:
            if self._positions_alloc:  # already allocated
                # Setting this will always zero fill position data
                # ie
                # True -> False -> True will wipe data from first True state
                cnp.PyArray_FILLWBYTE(self._pos, 0)
                self._has_positions = True
            else:  # first time, we need to allocate to correct shape
                self._pos = cnp.PyArray_ZEROS(
                    2, self._particle_dependent_dim, self._typenum, 0)
                self._has_positions = True
                self._positions_alloc = True
        elif not val:
            # Unsetting val won't delete the numpy array
            self._has_positions = False

    @property
    def has_velocities(self):
        """A boolean of whether this Timestep has velocity data
        This can be changed to ``True`` or ``False`` to allocate space for
        or remove the data.

        .. versionadded:: 0.11.0
        """
        return self._has_velocities

    @has_velocities.setter
    def has_velocities(self, val):
        if val and not self._has_velocities:
            if self._velocities_alloc:  # already allocated
                # Setting this will always zero fill velocity data
                # ie
                # True -> False -> True will wipe data from first True state
                cnp.PyArray_FILLWBYTE(self._velocities, 0)
                self._has_velocities = True
            else:  # first time, we need to allocate to correct shape
                self._velocities = cnp.PyArray_ZEROS(
                    2, self._particle_dependent_dim, self._typenum, 0)
                self._has_velocities = True
                self._velocities_alloc = True
        elif not val:
            # Unsetting val won't delete the numpy array
            self._has_velocities = False

    @property
    def has_forces(self):
        """A boolean of whether this Timestep has force data
        This can be changed to ``True`` or ``False`` to allocate space for
        or remove the data.

        .. versionadded:: 0.11.0
        """
        return self._has_forces

    @has_forces.setter
    def has_forces(self, val):
        if val and not self._has_forces:
            if self._forces_alloc:  # already allocated
                # Setting this will always zero fill force data
                # ie
                # True -> False -> True will wipe data from first True state
                cnp.PyArray_FILLWBYTE(self._forces, 0)
                self._has_forces = True
            else:  # first time, we need to allocate to correct shape
                self._forces = cnp.PyArray_ZEROS(
                    2, self._particle_dependent_dim, self._typenum, 0)
                self._has_forces = True
                self._forces_alloc = True
        elif not val:
            # Unsetting val won't delete the numpy array
            self._has_forces = False

    @property
    def positions(self):
        """A record of the positions of all atoms in this Timestep
        Setting this attribute will add positions to the Timestep if they
        weren't originally present.

        Returns
        -------
        positions : numpy.ndarray with dtype numpy.float32
               position data of shape ``(n_atoms, 3)`` for all atoms

        Raises
        ------
        :exc:`MDAnalysis.exceptions.NoDataError`
               if the Timestep has no position data


        .. versionchanged:: 0.11.0
           Now can raise :exc:`NoDataError` when no position data present
        """
        if self._has_positions:
            return self._pos
        else:
            raise NoDataError("This Timestep has no position information")

    @positions.setter
    def positions(self,  new_positions):
        self.has_positions = True
        if cnp.PyArray_Check(new_positions):  # is it an array?
            cnp.PyArray_CopyInto(self._pos, new_positions)
            # copy into target, handles dtype conversion
        else:
            self._pos[:] = new_positions

    @property
    def _x(self):
        """A view onto the x dimension of position data

        .. versionchanged:: 0.11.0
           Now read only
        """
        return self.positions[:, 0]

    @property
    def _y(self):
        """A view onto the y dimension of position data

        .. versionchanged:: 0.11.0
           Now read only
        """
        return self.positions[:, 1]

    @property
    def _z(self):
        """A view onto the z dimension of position data

        .. versionchanged:: 0.11.0
           Now read only
        """
        return self.positions[:, 2]

    @property
    def dimensions(self):
        """View of unitcell dimensions (*A*, *B*, *C*, *alpha*, *beta*, *gamma*)

        lengths *a*, *b*, *c* are in the MDAnalysis length unit (Å), and
        angles are in degrees.
        """
        if (self._unitcell[:3] == 0).all():
            return None
        else:
            return self._unitcell

    @dimensions.setter
    def dimensions(self,  new_dimensions):
        if new_dimensions is None:
            self._unitcell[:] = 0
        else:
            self._unitcell[:] = new_dimensions

    @property
    def volume(self):
        """volume of the unitcell"""
        if self.dimensions is None:
            return 0
        else:
            return core.box_volume(self.dimensions)

    @property
    def triclinic_dimensions(self):
        """The unitcell dimensions represented as triclinic vectors

        Returns
        -------
        numpy.ndarray
             A (3, 3) numpy.ndarray of unit cell vectors

        Examples
        --------
        The unitcell for a given system can be queried as either three
        vectors lengths followed by their respective angle, or as three
        triclinic vectors.

          >>> ts.dimensions
          array([ 13.,  14.,  15.,  90.,  90.,  90.], dtype=float32)
          >>> ts.triclinic_dimensions
          array([[ 13.,   0.,   0.],
                 [  0.,  14.,   0.],
                 [  0.,   0.,  15.]], dtype=float32)

        Setting the attribute also works::

          >>> ts.triclinic_dimensions = [[15, 0, 0], [5, 15, 0], [5, 5, 15]]
          >>> ts.dimensions
          array([ 15.        ,  15.81138802,  16.58312416,  67.58049774,
                  72.45159912,  71.56504822], dtype=float32)

        See Also
        --------
        :func:`MDAnalysis.lib.mdamath.triclinic_vectors`


        .. versionadded:: 0.11.0
        """
        if self.dimensions is None:
            return None
        else:
            return core.triclinic_vectors(self.dimensions)

    @triclinic_dimensions.setter
    def triclinic_dimensions(self, new_dimensions):
        """Set the unitcell for this Timestep as defined by triclinic vectors
        .. versionadded:: 0.11.0
        """
        if new_dimensions is None:
            self.dimensions = None
        else:
            self.dimensions = core.triclinic_box(*new_dimensions)

    @property
    def velocities(self):
        """A record of the velocities of all atoms in this Timestep
        Setting this attribute will add velocities to the Timestep if they
        weren't originally present.

        Returns
        -------
        velocities : numpy.ndarray with dtype numpy.float32
               velocity data of shape ``(n_atoms, 3)`` for all atoms

        Raises
        ------
        :exc:`MDAnalysis.exceptions.NoDataError`
               if the Timestep has no velocity data


        .. versionadded:: 0.11.0
        """
        if self._has_velocities:
            return self._velocities
        else:
            raise NoDataError("This Timestep has no velocities information")

    @velocities.setter
    def velocities(self,  new_velocities):
        self.has_velocities = True
        if cnp.PyArray_Check(new_velocities):  # is it an array?
            cnp.PyArray_CopyInto(self._velocities, new_velocities)
            # copy into target, handles dtype conversion
        else:
            self._velocities[:] = new_velocities

    @property
    def forces(self):
        """A record of the forces of all atoms in this Timestep
        Setting this attribute will add forces to the Timestep if they
        weren't originally present.

        Returns
        -------

        forces : numpy.ndarray with dtype numpy.float32
               force data of shape ``(n_atoms, 3)`` for all atoms
        Raises
        ------

        :exc:`MDAnalysis.exceptions.NoDataError`
               if the Timestep has no force data
        .. versionadded:: 0.11.0
        """
        if self._has_forces:
            return self._forces
        else:
            raise NoDataError("This Timestep has no force information")

    @forces.setter
    def forces(self,  new_forces):
        self.has_forces = True
        if cnp.PyArray_Check(new_forces):  # is it an array?
            cnp.PyArray_CopyInto(self._forces, new_forces)
            # copy into target, handles dtype conversion
        else:
            self._forces[:] = new_forces

    @classmethod
    def from_timestep(cls, Timestep other, **kwargs):
        """Create a copy of another Timestep, in the format of this Timestep

        .. versionadded:: 0.11.0
        """
        ts = cls(other.n_atoms,
                 positions=other.has_positions,
                 velocities=other.has_velocities,
                 forces=other.has_forces,
                 **kwargs)
        ts.frame = other.frame
        if other.dimensions is not None:
            ts.dimensions = other.dimensions.copy(order=cls.order)
        try:
            ts.positions = other.positions.copy(order=cls.order)
        except NoDataError:
            pass
        try:
            ts.velocities = other.velocities.copy(order=cls.order)
        except NoDataError:
            pass
        try:
            ts.forces = other.forces.copy(order=cls.order)
        except NoDataError:
            pass

        try:
            other._reader = weakref.ref(ts._reader())
        except TypeError:  # TypeError from calling None weakref
            pass

        ts.data = copy.deepcopy(other.data)

        return ts

    @classmethod
    def from_coordinates(cls,
                         positions=None,
                         velocities=None,
                         forces=None,
                         **kwargs):
        """Create an instance of this Timestep, from coordinate data

        Can pass position, velocity and force data to form a Timestep.

        .. versionadded:: 0.11.0
        """
        has_positions = positions is not None
        has_velocities = velocities is not None
        has_forces = forces is not None

        lens = [len(a) for a in [positions, velocities, forces]
                if a is not None]
        if not lens:
            raise ValueError("Must specify at least one set of data")
        n_atoms = max(lens)
        # Check arrays are matched length?
        if not all(val == n_atoms for val in lens):
            raise ValueError("Lengths of input data mismatched")

        ts = cls(n_atoms,
                 positions=has_positions,
                 velocities=has_velocities,
                 forces=has_forces,
                 **kwargs)
        if has_positions:
            ts.positions = positions
        if has_velocities:
            ts.velocities = velocities
        if has_forces:
            ts.forces = forces

        return ts

    def __eq__(self, other):
        """Compare with another Timestep

        .. versionadded:: 0.11.0
        """
        if not isinstance(other, Timestep):
            return NotImplemented

        if not self.frame == other.frame:
            return False

        if not self.n_atoms == other.n_atoms:
            return False

        if not self.has_positions == other.has_positions:
            return False
        if self.has_positions:
            if not (self.positions == other.positions).all():
                return False

        if self.dimensions is None:
            if other.dimensions is not None:
                return False
        else:
            if other.dimensions is None:
                return False
            if not (self.dimensions == other.dimensions).all():
                return False

        if not self.has_velocities == other.has_velocities:
            return False
        if self.has_velocities:
            if not (self.velocities == other.velocities).all():
                return False

        if not self.has_forces == other.has_forces:
            return False
        if self.has_forces:
            if not (self.forces == other.forces).all():
                return False

        return True

    # __ne__ is defers to __eq__ and inverts

    def __getitem__(self, atoms):
        """Get a selection of coordinates

        ``ts[i]``

           return coordinates for the i'th atom (0-based)

        ``ts[start:stop:skip]``

           return an array of coordinates, where start, stop and skip
           correspond to atom indices,
           :attr:`MDAnalysis.core.groups.Atom.index` (0-based)
        """
        if isinstance(atoms, numbers.Integral):
            return self._pos[atoms]
        elif isinstance(atoms, (slice, cnp.ndarray)):
            return self._pos[atoms]
        else:
            raise TypeError

    def __getattr__(self, attr):
        # special-case timestep info
        if attr in ('velocities', 'forces', 'positions'):
            raise NoDataError('This Timestep has no ' + attr)
        err = "{selfcls} object has no attribute '{attr}'"
        raise AttributeError(err.format(selfcls=type(self).__name__,
                                        attr=attr))

    def __len__(self):
        return self.n_atoms

    def __iter__(self):
        """Iterate over coordinates

        ``for x in ts``

        iterate of the coordinates, atom by atom
        """
        for i in range(self.n_atoms):
            yield self[i]

    def __repr__(self):
        desc = "< Timestep {0}".format(self.frame)
        if self.dimensions is not None:
            tail = " with unit cell dimensions {0} >".format(self.dimensions)
        else:
            tail = " >"
        return desc + tail

    def copy(self):
        """Make an independent ("deep") copy of the whole :class:`Timestep`."""
        return self.__deepcopy__()

    def __deepcopy__(self):
        return self.from_timestep(self)

    def __getstate__(self):
        """Make a dictionary of the class state to pickle Timestep instance.

           Must be done manually as Extension types do not have the __dict__
           class attribute and we use a non-trivial `__cinit__`. This means
           that cython cannot automatically  generate an `__reduce__` method
           for us.

        .. versionadded:: 2.3.0
        """
        state = {
            "frame": self.frame,
            "_n_atoms": self._n_atoms,
            "_has_positions": self._has_positions,
            "_has_velocities": self._has_velocities,
            "_has_forces": self._has_forces,

            "_unitcell": self._unitcell,
            "_pos": self._pos,
            "_velocities": self._velocities,
            "_forces": self._forces,

            "_dtype": self._dtype,
            "data": self.data,
            "aux": self.aux,
            "dt": self.dt
        }
        return state

    def __getnewargs_ex__(self):
        """Specify arguments to use in `__cinit__` and `__init__` to use in
           unpickling of timestep instance

        .. versionchanged:: 2.3.0
           removed implementations that use `__dict__` class attribute

        """
        return (self.n_atoms,), {}

    def __setstate__(self, state):
        """Restore class from `state` dictionary in unpickling of Timestep
           instance

        .. versionchanged:: 2.3.0
           removed implementations that use `__dict__` class attribute

        """
        self.frame = state["frame"]
        self._n_atoms = state["_n_atoms"]
        self.has_positions = state["_has_positions"]
        self._has_velocities = state["_has_velocities"]
        self._has_forces = state["_has_forces"]
        self._unitcell = state["_unitcell"]
        self._pos = state["_pos"]
        self._velocities = state["_velocities"]
        self._forces = state["_forces"]
        self._dtype = state["_dtype"]
        self.data = state["data"]
        self.aux = state["aux"]

    def copy_slice(self, sel):
        """Make a new `Timestep` containing a subset of the original `Timestep`.

        Parameters
        ----------
        sel : array_like or slice
            The underlying position, velocity, and force arrays are sliced
            using a :class:`list`, :class:`slice`, or any array-like.

        Returns
        -------
        :class:`Timestep`
            A `Timestep` object of the same type containing all header
            information and all atom information relevant to the selection.

        Note
        ----
        The selection must be a 0 based :class:`slice` or array of the atom indices
        in this :class:`Timestep`

        Example
        -------
        Using a Python :class:`slice` object::

           new_ts = ts.copy_slice(slice(start, stop, step))

        Using a list of indices::

           new_ts = ts.copy_slice([0, 2, 10, 20, 23])


        .. versionadded:: 0.8
        .. versionchanged:: 0.11.0
           Reworked to follow new Timestep API.  Now will strictly only
           copy official attributes of the Timestep.

        """
        # Detect the size of the Timestep by doing a dummy slice
        try:
            pos = self.positions[sel, :]
        except NoDataError:
            # It's cool if there's no Data, we'll live
            pos = None
        except Exception:
            errmsg = ("Selection type must be compatible with slicing the "
                      "coordinates")
            raise TypeError(errmsg) from None
        try:
            vel = self.velocities[sel, :]
        except NoDataError:
            vel = None
        except Exception:
            errmsg = ("Selection type must be compatible with slicing the "
                      "coordinates")
            raise TypeError(errmsg) from None
        try:
            force = self.forces[sel, :]
        except NoDataError:
            force = None
        except Exception:
            errmsg = ("Selection type must be compatible with slicing the "
                      "coordinates")
            raise TypeError(errmsg) from None

        new_TS = self.__class__.from_coordinates(
            positions=pos,
            velocities=vel,
            forces=force)

        new_TS.dimensions = self.dimensions

        new_TS.frame = self.frame

        try:
            new_TS._reader = weakref.ref(self._reader())
        except TypeError:  # TypeError from calling None weakref
            pass
        new_TS.data = copy.deepcopy(self.data)

        return new_TS

    @property
    def dt(self):
        """The time difference in ps between timesteps

        Note
        ----
        This defaults to 1.0 ps in the absence of time data


        .. versionadded:: 0.11.0
        """
        try:
            return self.data['dt']
        except KeyError:
            pass
        try:
            dt = self.data['dt'] = self._reader()._get_dt()
            return dt
        # TypeError from calling None weakref
        # AttributeError from ._get_dt()
        except (TypeError, AttributeError):
            pass

        warnings.warn("Reader has no dt information, set to 1.0 ps")
        return 1.0

    @dt.setter
    def dt(self, new_dt):
        self.data['dt'] = new_dt

    @dt.deleter
    def dt(self):
        del self.data['dt']

    @property
    def time(self):
        """The time in ps of this timestep

        This is calculated as::

          time = ts.data['time_offset'] + ts.time

        Or, if the trajectory doesn't provide time information::

          time = ts.data['time_offset'] + ts.frame * ts.dt

        .. versionadded:: 0.11.0
        """
        offset = self.data.get('time_offset', 0)
        try:
            return self.data['time'] + offset
        except KeyError:
            return self.dt * self.frame + offset

    @time.setter
    def time(self, new_time):
        self.data['time'] = new_time

    @time.deleter
    def time(self):
        del self.data['time']



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pytest

from importlib import reload
import pytest

from MDAnalysis.analysis import hole2
from MDAnalysisTests.util import import_not_available


@pytest.mark.skipif(
    import_not_available("mdahole2"),
    reason="Test skipped because mdahole2 is not found",
)
def test_moved_to_mdakit_warning():
    wmsg = "MDAnalysis.analysis.hole2 is deprecated"
    with pytest.warns(DeprecationWarning, match=wmsg):
        reload(hole2)


@pytest.mark.skipif(
    not import_not_available("mdahole2"),
    reason="Test skipped because mdahole2 is found",
)
def test_install_mdakit_warning():
    wmsg = "Please install"
    with pytest.warns(UserWarning, match=wmsg):
        reload(hole2)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
from MDAnalysisTests.topology.base import ParserBase
from MDAnalysisTests.datafiles import DMS_DOMAINS, DMS_NO_SEGID


class TestDMSParser(ParserBase):
    parser = mda.topology.DMSParser.DMSParser
    ref_filename = DMS_DOMAINS
    expected_attrs = [
        "ids",
        "names",
        "bonds",
        "charges",
        "masses",
        "resids",
        "resnames",
        "segids",
        "chainIDs",
        "atomnums",
    ]
    guessed_attrs = ["types"]
    expected_n_atoms = 3341
    expected_n_residues = 214
    expected_n_segments = 3

    def test_number_of_bonds(self, top):
        assert len(top.bonds.values) == 3365

    def test_atomsels(self, filename):
        # Desired value taken from VMD atomsel
        u = mda.Universe(filename)

        s0 = u.select_atoms("name CA")
        assert len(s0) == 214

        s1 = u.select_atoms("resid 33")
        assert len(s1) == 12

        s2 = u.select_atoms("segid NMP")
        assert len(s2) == 437

        s3 = u.select_atoms("segid LID")
        assert len(s3) == 598

        s4 = u.select_atoms("segid CORE")
        assert len(s4) == 2306

        s5 = u.select_atoms("resname ALA")
        assert len(s5) == 190

    def test_guessed_types(self, filename):
        u = mda.Universe(filename)
        expected = ["N", "H", "H", "H", "C", "H", "C"]
        assert (u.atoms.types[:7] == expected).all()


class TestDMSParserNoSegid(TestDMSParser):
    ref_filename = DMS_NO_SEGID
    expected_n_segments = 1

    def test_atomsels(self, filename):
        u = mda.Universe(filename)

        s0 = u.select_atoms("name CA")
        assert len(s0) == 214

        s1 = u.select_atoms("resid 33")
        assert len(s1) == 12

        s2 = u.select_atoms("segid SYSTEM")
        assert len(s2) == 3341

        s5 = u.select_atoms("resname ALA")
        assert len(s5) == 190



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Neighbor Search wrapper for MDAnalysis --- :mod:`MDAnalysis.lib.NeighborSearch`
===============================================================================

This module contains classes that allow neighbor searches directly with
`AtomGroup` objects from `MDAnalysis`.
"""
import numpy as np
from MDAnalysis.lib.distances import capped_distance
from MDAnalysis.lib.util import unique_int_1d
from MDAnalysis.core.groups import AtomGroup, SegmentGroup, ResidueGroup
import numpy.typing as npt
from typing import Optional, Union, List


class AtomNeighborSearch(object):
    """This class can be used to find all atoms/residues/segments within the
    radius of a given query position.

    For the neighbor search, this class is a wrapper around
    :class:`~MDAnalysis.lib.distances.capped_distance`.
    """

    def __init__(
        self, atom_group: AtomGroup, box: Optional[npt.ArrayLike] = None
    ) -> None:
        """

        Parameters
        ----------
        atom_list : AtomGroup
          list of atoms
        box : array-like or ``None``, optional, default ``None``
          Simulation cell dimensions in the form of
          :attr:`MDAnalysis.trajectory.timestep.Timestep.dimensions` when
          periodic boundary conditions should be taken into account for
          the calculation of contacts.
        """
        self.atom_group = atom_group
        self._u = atom_group.universe
        self._box = box

    def search(
        self, atoms: AtomGroup, radius: float, level: str = "A"
    ) -> Optional[Union[AtomGroup, ResidueGroup, SegmentGroup]]:
        """
        Return all atoms/residues/segments that are within *radius* of the
        atoms in *atoms*.

        Parameters
        ----------
        atoms : AtomGroup, MDAnalysis.core.groups.AtomGroup
          AtomGroup object
        radius : float
          Radius for search in Angstrom.
        level : str
          char (A, R, S). Return atoms(A), residues(R) or segments(S) within
          *radius* of *atoms*.

        Returns
        -------
        AtomGroup : :class:`~MDAnalysis.core.groups.AtomGroup`
          When ``level='A'``, AtomGroup is being returned.
        ResidueGroup : :class:`~MDAnalysis.core.groups.ResidueGroup`
          When ``level='R'``, ResidueGroup is being returned.
        SegmentGroup : :class:`~MDAnalysis.core.groups.SegmentGroup`
          When ``level='S'``, SegmentGroup is being returned.


        .. versionchanged:: 2.0.0
           Now returns :class:`AtomGroup` (when empty this is now an empty
           :class:`AtomGroup` instead of an empty list), :class:`ResidueGroup`,
           or a :class:`SegmentGroup`
        """
        unique_idx = []
        try:
            # For atom groups, take the positions attribute
            position = atoms.positions
        except AttributeError:
            # For atom, take the position attribute
            position = atoms.position
        pairs = capped_distance(
            position,
            self.atom_group.positions,
            radius,
            box=self._box,
            return_distances=False,
        )

        if pairs.size > 0:
            unique_idx = unique_int_1d(np.asarray(pairs[:, 1], dtype=np.intp))
        return self._index2level(unique_idx, level)

    def _index2level(
        self, indices: List[int], level: str
    ) -> Union[AtomGroup, ResidueGroup, SegmentGroup]:
        """Convert list of atom_indices in a AtomGroup to either the
        Atoms or segments/residues containing these atoms.

        Parameters
        ----------
        indices
           list of atom indices
        level : str
          char (A, R, S). Return atoms(A), residues(R) or segments(S) within
          *radius* of *atoms*.
        """
        atomgroup = self.atom_group[indices]
        if level == "A":
            return atomgroup
        elif level == "R":
            return atomgroup.residues
        elif level == "S":
            return atomgroup.segments
        else:
            raise NotImplementedError(
                "{0}: level not implemented".format(level)
            )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
PyMOL selections
=================

Write :class:`MDAnalysis.core.groups.AtomGroup` selection to a
script `pml`_ file that defines PyMOL_ atomselect macros. To be used
in PyMOL like this::

  @macros.pml

The selections should appear in the user interface.

.. _PyMOL: http://www.pymol.org
.. _pml: http://pymol.sourceforge.net/newman/user/S0210start_cmds.html#6_5_1

.. autoclass:: SelectionWriter
   :inherited-members:
"""
from . import base


class SelectionWriter(base.SelectionWriterBase):
    format = ["PyMol", "pml"]
    ext = "pml"
    continuation = "\\"  # quoted backslash!
    commentfmt = "# %s"
    default_numterms = 6

    def _translate(self, atoms, **kwargs):
        # PyMol index is 1-based
        def _index(atom):
            return "index {0:d}".format((atom.index + 1))

        return base.join(atoms, " |", _index)

    def _write_head(self, out, **kwargs):
        out.write(self.comment("MDAnalysis PyMol selection"))
        out.write(
            "select {name!s}, ".format(**kwargs) + self.continuation + "\n"
        )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
bootstrap procedures --- :mod:`MDAnalysis.analysis.ensemble.bootstrap`
======================================================================


The module contains functions for bootstrapping either ensembles (Universe
objects) or distance matrices, by resampling with replacement.

:Author: Matteo Tiberti, Wouter Boomsma, Tone Bengtsen

.. versionadded:: 0.16.0

.. deprecated:: 2.8.0
   This module is deprecated in favour of the 
   MDAKit `mdaencore <https://mdanalysis.org/mdaencore/>`_ and will be removed
   in MDAnalysis 3.0.0.

"""
import numpy as np
import logging
import MDAnalysis as mda
from .utils import TriangularMatrix, ParallelCalculation


def bootstrapped_matrix(matrix, ensemble_assignment):
    """
    Bootstrap an input square matrix. The resulting matrix will have the same
    shape as the original one, but the order of its elements will be drawn
    (with repetition). Separately bootstraps each ensemble.

    Parameters
    ----------

    matrix : encore.utils.TriangularMatrix
        similarity/dissimilarity matrix

    ensemble_assignment: numpy.array
        array of ensemble assignments. This array must be matrix.size long.

    Returns
    -------

    this_m : encore.utils.TriangularMatrix
        bootstrapped similarity/dissimilarity matrix
    """
    ensemble_identifiers = np.unique(ensemble_assignment)
    this_m = TriangularMatrix(size=matrix.size)
    indexes = []
    for ens in ensemble_identifiers:
        old_indexes = np.where(ensemble_assignment == ens)[0]
        indexes.append(
            np.random.randint(
                low=np.min(old_indexes),
                high=np.max(old_indexes) + 1,
                size=old_indexes.shape[0],
            )
        )

    indexes = np.hstack(indexes)
    for j in range(this_m.size):
        for k in range(j):
            this_m[j, k] = matrix[indexes[j], indexes[k]]

    logging.info("Matrix bootstrapped.")
    return this_m


def get_distance_matrix_bootstrap_samples(
    distance_matrix, ensemble_assignment, samples=100, ncores=1
):
    """
    Calculates distance matrices corresponding to bootstrapped ensembles, by
    resampling with replacement.

    Parameters
    ----------

    distance_matrix : encore.utils.TriangularMatrix
        Conformational distance matrix

    ensemble_assignment : str
        Mapping from frames to which ensemble they are from (necessary because
        ensembles are bootstrapped independently)

    samples : int, optional
        How many bootstrap samples to create.

    ncores : int, optional
        Maximum number of cores to be used (default is 1)

    Returns
    -------

    confdistmatrix : list of encore.utils.TriangularMatrix
    """

    bs_args = [
        ([distance_matrix, ensemble_assignment]) for i in range(samples)
    ]

    pc = ParallelCalculation(ncores, bootstrapped_matrix, bs_args)

    pc_results = pc.run()

    bootstrap_matrices = list(zip(*pc_results))[1]

    return bootstrap_matrices


def get_ensemble_bootstrap_samples(ensemble, samples=100):
    """
    Generates a bootstrapped ensemble by resampling with replacement.

    Parameters
    ----------

    ensemble : MDAnalysis.Universe
        Conformational distance matrix

    samples : int, optional
        How many bootstrap samples to create.

    Returns
    -------

    list of MDAnalysis.Universe objects
    """

    ensemble.transfer_to_memory()

    ensembles = []
    for i in range(samples):
        indices = np.random.randint(
            low=0,
            high=ensemble.trajectory.timeseries().shape[1],
            size=ensemble.trajectory.timeseries().shape[1],
        )
        ensembles.append(
            mda.Universe(
                ensemble.filename,
                ensemble.trajectory.timeseries(order="fac")[indices, :, :],
                format=mda.coordinates.memory.MemoryReader,
            )
        )
    return ensembles



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import os
from pathlib import Path
import numpy as np

import MDAnalysis as mda
from MDAnalysis.coordinates.DCD import DCDReader

from numpy.testing import (
    assert_equal,
    assert_array_equal,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_allclose,
)

from MDAnalysisTests.datafiles import (
    DCD,
    PSF,
    DCD_empty,
    PRMncdf,
    NCDF,
    COORDINATES_TOPOLOGY,
    COORDINATES_DCD,
    PSF_TRICLINIC,
    DCD_TRICLINIC,
    PSF_NAMD_TRICLINIC,
    DCD_NAMD_TRICLINIC,
    PSF_NAMD_GBIS,
    DCD_NAMD_GBIS,
    PDB_closed,
)
from MDAnalysisTests.coordinates.base import (
    MultiframeReaderTest,
    BaseReference,
    BaseWriterTest,
)

import pytest


class DCDReference(BaseReference):
    def __init__(self):
        super(DCDReference, self).__init__()
        self.trajectory = COORDINATES_DCD
        self.topology = COORDINATES_TOPOLOGY
        self.reader = mda.coordinates.DCD.DCDReader
        self.writer = mda.coordinates.DCD.DCDWriter
        self.ext = "xtc"
        self.prec = 3
        self.changing_dimensions = True


class TestDCDReader(MultiframeReaderTest):
    @staticmethod
    @pytest.fixture()
    def ref():
        return DCDReference()

    def test_empty_dcd(self):
        with pytest.raises(IOError):
            mda.Universe(PSF, DCD_empty)

    def test_with_statement(self):

        try:
            with DCDReader(DCD) as trj:
                N = trj.n_frames
                frames = [ts.frame for ts in trj]
        except:
            raise AssertionError("with_statement not working for DCDReader")
        assert_equal(
            N,
            98,
            err_msg="with_statement: DCDReader reads wrong number of frames",
        )
        assert_array_equal(
            frames,
            np.arange(0, N),
            err_msg="with_statement: DCDReader does not read all frames",
        )

    def test_set_time(self):
        u = mda.Universe(PSF, DCD)
        assert_almost_equal(u.trajectory.time, 1.0, decimal=5)


@pytest.mark.parametrize("fstart", (0, 1, 2, 37, None))
def test_write_istart(universe_dcd, tmpdir, fstart):
    outfile = str(tmpdir.join("test.dcd"))
    nsavc = universe_dcd.trajectory._file.header["nsavc"]
    istart = fstart * nsavc if fstart is not None else None
    with mda.Writer(
        outfile, universe_dcd.atoms.n_atoms, nsavc=nsavc, istart=istart
    ) as w:
        for ts in universe_dcd.trajectory:
            w.write(universe_dcd.atoms)
    u = mda.Universe(PSF, outfile)
    assert_almost_equal(
        u.trajectory._file.header["istart"],
        istart if istart is not None else u.trajectory._file.header["nsavc"],
    )
    # issue #1819
    times = [ts.time for ts in u.trajectory]

    fstart = fstart if fstart is not None else 1
    ref_times = (
        np.arange(universe_dcd.trajectory.n_frames) + fstart
    ) * universe_dcd.trajectory.dt
    assert_almost_equal(
        times,
        ref_times,
        decimal=5,
        err_msg="Times not identical after setting istart={}".format(istart),
    )


class TestDCDWriter(BaseWriterTest):
    @staticmethod
    @pytest.fixture()
    def ref():
        return DCDReference()


def test_write_random_unitcell(tmpdir):
    testname = str(tmpdir.join("test.dcd"))
    rstate = np.random.RandomState(1178083)
    random_unitcells = rstate.uniform(low=5, high=120, size=(98, 6)).astype(
        np.float64
    )

    u = mda.Universe(PSF, DCD)
    with mda.Writer(testname, n_atoms=u.atoms.n_atoms) as w:
        for index, ts in enumerate(u.trajectory):
            u.atoms.dimensions = random_unitcells[index]
            w.write(u.atoms)

    u2 = mda.Universe(PSF, testname)
    for index, ts in enumerate(u2.trajectory):
        assert_allclose(
            u2.trajectory.dimensions,
            random_unitcells[index],
            rtol=1e-5,
        )


def test_empty_dimension_warning(tmpdir):

    u = mda.Universe(PDB_closed)
    testname = str(tmpdir.join("test.dcd"))

    with mda.Writer(testname, n_atoms=u.atoms.n_atoms) as w:
        msg = "zeroed unitcell will be written"
        with pytest.warns(UserWarning, match=msg):
            w.write(u.atoms)


################
# Legacy tests #
################


@pytest.fixture(scope="module")
def universe_dcd():
    return mda.Universe(PSF, DCD)


def test_rewind(universe_dcd):
    universe_dcd.trajectory.rewind()
    assert universe_dcd.trajectory.ts.frame == 0


def test_next(universe_dcd):
    universe_dcd.trajectory.rewind()
    universe_dcd.trajectory.next()
    assert universe_dcd.trajectory.ts.frame == 1


def test_jump_last_frame(universe_dcd):
    universe_dcd.trajectory[-1]
    assert universe_dcd.trajectory.ts.frame == 97


@pytest.mark.parametrize("start, stop, step", ((5, 17, 3), (20, 5, -1)))
def test_slice(universe_dcd, start, stop, step):
    frames = [ts.frame for ts in universe_dcd.trajectory[start:stop:step]]
    assert_array_equal(frames, np.arange(start, stop, step))


@pytest.mark.parametrize("array_like", [list, np.array])
def test_array_like(universe_dcd, array_like):
    ar = array_like([0, 3, 4, 5])
    frames = [ts.frame for ts in universe_dcd.trajectory[ar]]
    assert_array_equal(frames, ar)


@pytest.mark.parametrize(
    "indices", ([0, 4, 2, 3, 0, 1], [0, 0, 1, 1, 2, 1, 1])
)
def test_list_indices(universe_dcd, indices):
    frames = [ts.frame for ts in universe_dcd.trajectory[indices]]
    assert_array_equal(frames, indices)


@pytest.mark.parametrize(
    "slice, length",
    [
        ([None, None, None], 98),
        ([0, None, None], 98),
        ([None, 98, None], 98),
        ([None, None, 1], 98),
        ([None, None, -1], 98),
        ([2, 6, 2], 2),
        ([0, 10, None], 10),
        ([2, 10, None], 8),
        ([0, 1, 1], 1),
        ([1, 1, 1], 0),
        ([1, 2, 1], 1),
        ([1, 2, 2], 1),
        ([1, 4, 2], 2),
        ([1, 4, 4], 1),
        ([0, 5, 5], 1),
        ([3, 5, 1], 2),
        ([4, 0, -1], 4),
        ([5, 0, -2], 3),
        ([5, 0, -4], 2),
    ],
)
def test_timeseries_slices(slice, length, universe_dcd):
    start, stop, step = slice
    allframes = universe_dcd.trajectory.timeseries(order="fac")
    xyz = universe_dcd.trajectory.timeseries(
        start=start, stop=stop, step=step, order="fac"
    )
    assert len(xyz) == length
    assert_array_almost_equal(xyz, allframes[start:stop:step])


@pytest.mark.parametrize(
    "order, shape",
    (
        ("fac", (98, 3341, 3)),
        ("fca", (98, 3, 3341)),
        ("afc", (3341, 98, 3)),
        ("acf", (3341, 3, 98)),
        ("caf", (3, 3341, 98)),
        ("cfa", (3, 98, 3341)),
    ),
)
def test_timeseries_order(order, shape, universe_dcd):
    x = universe_dcd.trajectory.timeseries(order=order)
    assert x.shape == shape


@pytest.mark.parametrize(
    "indices", [[1, 2, 3, 4], [5, 10, 15, 19], [9, 4, 2, 0, 50]]
)
def test_timeseries_atomindices(indices, universe_dcd):
    allframes = universe_dcd.trajectory.timeseries(order="afc")
    asel = universe_dcd.atoms[indices]
    xyz = universe_dcd.trajectory.timeseries(asel=asel, order="afc")
    assert len(xyz) == len(indices)
    assert_array_almost_equal(xyz, allframes[indices])


def test_reader_set_dt():
    dt = 100
    frame = 3
    u = mda.Universe(PSF, DCD, dt=dt)
    dcdheader = u.trajectory._file.header
    fstart = dcdheader["istart"] / dcdheader["nsavc"]
    assert_almost_equal(
        u.trajectory[frame].time,
        (frame + fstart) * dt,
        err_msg="setting time step dt={0} failed: "
        "actually used dt={1}".format(dt, u.trajectory._ts_kwargs["dt"]),
    )
    assert_almost_equal(
        u.trajectory.dt, dt, err_msg="trajectory.dt does not match set dt"
    )


def test_writer_dt(tmpdir):
    dt = 5.0  # set time step to 5 ps for the written trajectory
    universe_dcd = mda.Universe(PSF, DCD)
    t = universe_dcd.trajectory
    outfile = str(tmpdir.join("test.dcd"))
    # use istart=None explicitly so that dcd start with time 1*dt
    # (DCD will set istart = nsavc)
    with mda.Writer(outfile, n_atoms=t.n_atoms, dt=dt, istart=None) as W:
        for ts in universe_dcd.trajectory:
            W.write(universe_dcd.atoms)

    uw = mda.Universe(PSF, outfile)

    # check that we actually changed dt
    assert not uw.trajectory.dt == pytest.approx(t.dt)

    assert_almost_equal(
        uw.trajectory.totaltime,
        (uw.trajectory.n_frames - 1) * dt,
        decimal=4,
        err_msg="Total time mismatch for DCD",
    )
    times = np.array([uw.trajectory.time for ts in uw.trajectory])
    frames = np.arange(1, uw.trajectory.n_frames + 1)  # traj starts at 1*dt
    assert_array_almost_equal(
        times,
        frames * dt,
        decimal=4,
        err_msg="Times mismatch for DCD",
    )


@pytest.mark.parametrize("variable, default", (("istart", 0), ("nsavc", 1)))
def test_DCDWriter_default(tmpdir, variable, default):
    outfile = str(tmpdir.join("test.dcd"))
    with mda.Writer(outfile, n_atoms=10) as W:
        pass  # just write empty trajectory with header
    # need to read because header might only be complete after close
    with mda.lib.formats.libdcd.DCDFile(outfile) as DCD:
        header = DCD.header
    assert header[variable] == default


@pytest.mark.parametrize("ext, decimal", (("dcd", 5), ("xtc", 2)))
def test_other_writer(universe_dcd, tmpdir, ext, decimal):
    t = universe_dcd.trajectory
    outfile = str(tmpdir.join("test.{}".format(ext)))
    with t.OtherWriter(outfile) as W:
        for ts in universe_dcd.trajectory:
            W.write(universe_dcd)

    uw = mda.Universe(PSF, outfile)
    # check that the coordinates are identical for each time step
    for orig_ts, written_ts in zip(universe_dcd.trajectory, uw.trajectory):
        assert_array_almost_equal(
            written_ts.positions,
            orig_ts.positions,
            decimal,
            err_msg="coordinate mismatch between "
            "original and written trajectory at "
            "frame {} (orig) vs {} (written)".format(
                orig_ts.frame, written_ts.frame
            ),
        )


def test_single_frame(universe_dcd, tmpdir):
    u = universe_dcd
    outfile = str(tmpdir.join("test.dcd"))
    with mda.Writer(outfile, u.atoms.n_atoms) as W:
        W.write(u.atoms)
    w = mda.Universe(PSF, outfile)
    assert w.trajectory.n_frames == 1
    assert_almost_equal(
        w.atoms.positions,
        u.atoms.positions,
        3,
        err_msg="coordinates do not match",
    )


def test_write_no_natoms():
    with pytest.raises(ValueError):
        mda.Writer("foobar.dcd")


def test_writer_trajectory_no_natoms(tmpdir, universe_dcd):
    with tmpdir.as_cwd():
        universe_dcd.trajectory.Writer("foo.dcd")


class RefCHARMMtriclinicDCD(object):
    topology = PSF_TRICLINIC
    trajectory = DCD_TRICLINIC
    # time(ps) A B C alpha beta gamma (length in Angstrome, angles in degrees)
    # dcd starts at t = 1ps
    ref_dimensions = np.array(
        [
            [1.0, 35.44604, 35.06156, 34.1585, 91.32802, 61.73521, 44.40703],
            [2.0, 34.65957, 34.22689, 33.09897, 90.56206, 61.79192, 44.14549],
            [3.0, 34.52772, 34.66422, 33.53881, 90.55859, 63.11228, 40.14044],
            [4.0, 34.43749, 33.38432, 34.02133, 88.82457, 64.98057, 36.77397],
            [5.0, 33.73129, 32.47752, 34.18961, 89.88102, 65.89032, 36.10921],
            [6.0, 33.78703, 31.90317, 34.98833, 90.03092, 66.12877, 35.07141],
            [7.0, 33.24708, 31.18271, 34.9654, 93.11122, 68.17743, 35.73643],
            [8.0, 32.92599, 30.31393, 34.99197, 93.89051, 69.3799, 33.48945],
            [9.0, 32.15295, 30.43056, 34.96157, 96.01416, 71.50115, 32.56111],
            [10.0, 31.99748, 30.21518, 35.24292, 95.85821, 71.08429, 31.85939],
        ]
    )


class RefNAMDtriclinicDCD(object):
    topology = PSF_NAMD_TRICLINIC
    trajectory = DCD_NAMD_TRICLINIC
    # vmd topology trajectory
    # molinfo 0 get {a b c alpha beta gamma}
    # time(ps) A B C alpha beta gamma (length in Angstrome, angles in degrees)
    ref_dimensions = np.array(
        [
            [
                1.0,
                38.426594,
                38.393101,
                44.759800,
                90.000000,
                90.000000,
                60.028915,
            ],
        ]
    )


@pytest.mark.parametrize("ref", (RefCHARMMtriclinicDCD, RefNAMDtriclinicDCD))
def test_read_unitcell_triclinic(ref):
    u = mda.Universe(ref.topology, ref.trajectory)
    for ts, box in zip(u.trajectory, ref.ref_dimensions[:, 1:]):
        assert_array_almost_equal(
            ts.dimensions,
            box,
            4,
            err_msg="box dimensions A,B,C,alpha,"
            "beta,gamma not identical at frame "
            "{}".format(ts.frame),
        )


@pytest.mark.parametrize("ref", (RefCHARMMtriclinicDCD, RefNAMDtriclinicDCD))
def test_write_unitcell_triclinic(ref, tmpdir):
    u = mda.Universe(ref.topology, ref.trajectory)
    outfile = "triclinic.dcd"
    with tmpdir.as_cwd():
        with u.trajectory.OtherWriter(outfile) as w:
            for ts in u.trajectory:
                w.write(u)

        w = mda.Universe(ref.topology, outfile)
        for ts_orig, ts_copy in zip(u.trajectory, w.trajectory):
            assert_almost_equal(
                ts_orig.dimensions,
                ts_copy.dimensions,
                4,
                err_msg="DCD->DCD: unit cell dimensions wrong "
                "at frame {0}".format(ts_orig.frame),
            )


@pytest.fixture(scope="module")
def ncdf2dcd(tmpdir_factory):
    pytest.importorskip("netCDF4")
    testfile = tmpdir_factory.mktemp("dcd").join("ncdf2dcd.dcd")
    testfile = str(testfile)
    ncdf = mda.Universe(PRMncdf, NCDF)
    with mda.Writer(testfile, n_atoms=ncdf.atoms.n_atoms) as w:
        for ts in ncdf.trajectory:
            w.write(ncdf)
    return ncdf, mda.Universe(PRMncdf, testfile)


def test_ncdf2dcd_unitcell(ncdf2dcd):
    ncdf, dcd = ncdf2dcd
    for ts_ncdf, ts_dcd in zip(ncdf.trajectory, dcd.trajectory):
        assert_almost_equal(ts_ncdf.dimensions, ts_dcd.dimensions, 3)


def test_ncdf2dcd_coords(ncdf2dcd):
    ncdf, dcd = ncdf2dcd
    for ts_ncdf, ts_dcd in zip(ncdf.trajectory, dcd.trajectory):
        assert_almost_equal(ts_ncdf.positions, ts_dcd.positions, 3)


@pytest.fixture(
    params=[
        (PSF, DCD),
        (PSF_TRICLINIC, DCD_TRICLINIC),
        (PSF_NAMD_TRICLINIC, DCD_NAMD_TRICLINIC),
        (PSF_NAMD_GBIS, DCD_NAMD_GBIS),
    ]
)
def universe(request):
    psf, dcd = request.param
    yield mda.Universe(psf, dcd)


def test_ts_time(universe):
    # issue #1819
    u = universe
    header = u.trajectory._file.header
    ref_times = [
        (ts.frame + header["istart"] / header["nsavc"]) * ts.dt
        for ts in u.trajectory
    ]
    times = [ts.time for ts in u.trajectory]
    assert_almost_equal(times, ref_times, decimal=5)


def test_pathlib():
    # regression test for DCD path of
    # gh-2497
    top = Path(PSF)
    traj = Path(DCD)
    u = mda.Universe(top, traj)
    # we really only care that pathlib
    # object handling worked
    assert u.atoms.n_atoms == 3341


@pytest.fixture
def large_dcdfile(tmpdir):
    # creates a >2Gb DCD file
    fsize = 3.8  # mb
    nreps_reqs = int(2100 // fsize)  # times to duplicate traj to hit 2.1Gb

    newf = str(tmpdir / "jabba.dcd")

    u = mda.Universe(PSF, DCD)

    with mda.Writer(newf, n_atoms=len(u.atoms)) as w:
        for _ in range(nreps_reqs):
            for ts in u.trajectory:
                w.write(u.atoms)

    yield newf, nreps_reqs


@pytest.mark.skipif(
    not os.environ.get("LARGEDCD", False), reason="Skipping large file test"
)
def test_large_dcdfile(large_dcdfile):
    DCD_large, nreps = large_dcdfile

    u_small = mda.Universe(PSF, DCD)
    u = mda.Universe(PSF, DCD_large)

    assert len(u.trajectory) == len(u_small.trajectory) * nreps

    u_small.trajectory[-1]
    u.trajectory[-1]

    assert_array_almost_equal(u.atoms.positions, u_small.atoms.positions)


def test_dcd_writer_angle_cosines(tmpdir):
    """Test that DCDWriter correctly converts angles to cosines as documented.

    This test verifies the fix for issue #5069 where DCDWriter was incorrectly
    writing angles in degrees instead of angle cosines.
    """
    u = mda.Universe(PSF_NAMD_TRICLINIC, DCD_NAMD_TRICLINIC)
    original_dimensions = u.dimensions.copy()

    outfile = str(tmpdir.join("test_angles.dcd"))
    u.atoms.write(outfile, frames="all")

    # Read back and check that dimensions are preserved
    u2 = mda.Universe(PSF_NAMD_TRICLINIC, outfile)

    # The dimensions should be identical when read back
    assert_allclose(
        u2.dimensions,
        original_dimensions,
        err_msg="DCDWriter failed to preserve unit cell dimensions (frame 0)",
    )

    # To get the originally stored dimensions, we need to directly access the DCD file.
    u.trajectory._reopen()
    frame = u.trajectory._file.read()
    frame2 = u2.trajectory._file.read()

    # direction cosines should be between -1 and 1
    assert np.all(
        np.abs(frame2.unitcell[[4, 3, 1]]) <= 1
    ), "DCDWriter failed to preserve unit cell angles as cosines"

    # Directly test the stored dimensions from the DCD header
    assert_allclose(
        frame2.unitcell,
        frame.unitcell,  # stored as cos(angle) in the test trajectory
        err_msg="DCDWriter failed to preserve unit cell dimensions as stored in the DCD file",
    )
